/*
 * VNPForth - Compiled native Forth for IA32 Linux systems
 * Copyright (C) 2001  Simon Baldwin (simonb@caldera.com)
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 */

#include <assert.h>
#include <stdio.h>
#include <sys/param.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>

#include "utils.h"
#include "ident.h"
#include "ifile.h"
#include "sfile.h"
#include "pstore.h"
#include "codegen.h"
#include "optimizer.h"
#include "getopts.h"
#include "forth_y.h"

/* Globally available data. */
char		*argv0;
extern int	yy_linenum, yy_errcount, yy_warncount;

/* Return status for the program. */
#define	COMPILE_SUCCESS		0
#define	COMPILE_FAIL		1


/*
 * compile()
 *
 * Compile the file already open as the current ifile.
 */
static int
compile (void)
{
	char	asmfile[MAXPATHLEN];		/* Assembler file name */
	char	pcodefile[MAXPATHLEN];		/* Intermediate file name */
	FILE	*pfile;				/* Intermediate file stream */
	char	objfile[MAXPATHLEN];		/* Object file name */
	char	ascmd[MAXPATHLEN];		/* Assembler command */
	int	status;				/* Record of success/failure */

	/* Compile the contents of the input file. */
	if (yy_parse_ifile () == 0 && yy_errcount == 0)
	    {
		/* Pass the optimizer over the pstore, if -O option used. */
		if (get_Oflag ())
			optimize ();

		/* Write out the pcode to <file>.p, if -P option used. */
		if (get_Pflag ())
		    {
			strcpy (pcodefile, get_ifname ());
			stripext (pcodefile);
			strcat (pcodefile, ".p");
			pfile = fopen (pcodefile, "w");
			if (pfile == NULL)
			    {
				fprintf (stderr, "%s: %s: ", argv0, pcodefile);
				perror ("");
				p_empty ();
				return COMPILE_FAIL;
			    }
			p_printout (pfile);
			fclose (pfile);
		    }

		/* Create the assembler output file <file>.s. */
		strcpy (asmfile, get_ifname ());
		stripext (asmfile);
		strcat (asmfile, ".s");
		if (open_sfile (asmfile) != 0)
		    {
			fprintf (stderr, "%s: %s: ", argv0, asmfile);
			perror ("");
			p_empty ();
			return COMPILE_FAIL;
		    }

		/* Generate native mode assembler and close output file. */
		generate_assembler (get_ifname ());
		close_sfile ();

		/* Empty out the pstore for the next file. */
		p_empty ();

		/* Complete the compilation with the assembler. */
		strcpy (objfile, get_ifname ());
		stripext (objfile);
		strcat (objfile, ".o");
		snprintf (ascmd, sizeof (ascmd), "as -o %s %s",
							objfile, asmfile);
		status = system (ascmd);
		if (status != 0)
		    {
			fprintf (stderr, "%s: %s: ", argv0, objfile);
			fprintf (stderr, "%s, status %d\n", ascmd, status);
			return COMPILE_FAIL;
		    }

		/* Remove the sfile unless -S option used. */
		if (!get_Sflag ())
			unlink_sfile ();

		/* Completely successful compilation. */
		return COMPILE_SUCCESS;
	    }
	else
	    {
		/* Unsuccessful parse - close and delete the sfile. */
		close_sfile ();
		unlink_sfile ();

		/* Empty out the pstore for the next file. */
		p_empty ();
		return COMPILE_FAIL;
	    }
}


/*
 * main()
 *
 * main program
 */
int
main (int argc, char *argv[])
{
	int	exit_status = COMPILE_SUCCESS;	/* Program exit status code */
	int	argnum;				/* Argument index */
	char	*argument;			/* Command line argument */

	/* Set up argv0 for everyone. */
	argv0 = basefilename (argv[0]);

	/* Parse options, then process each argument. */
	for (argnum = parse_cmdline (argc, argv); argnum < argc; argnum++)
	    {
		argument = argv[argnum];

		/*
		 * If no -M flag was given, the argument is a source file
		 * to compile.  If -M was used, the argument is an id that
		 * the user would like mangled.
		 */
		if (!get_Mflag ())
		    {
			/* Confirm the file exists. */
			if (open_ifile (argument) != 0)
			    {
				fprintf (stderr, "%s: %s: ", argv0, argument);
				perror ("");
				exit_status = COMPILE_FAIL;
				continue;
			    }

			/* Compile the file, and close it. */
			if (compile () != COMPILE_SUCCESS)
				exit_status = COMPILE_FAIL;
			close_ifile ();
		    }
		else
		    {
			/* Simply mangle the argument (id) and print it out. */
			printf ("%s\n", mangle (argument));
		    }
	    }

	/* Return zero unless any errors were encountered. */
	exit (exit_status);
}
