/*
 * VNPForth - Compiled native Forth for IA32 Linux systems
 * Copyright (C) 2001  Simon Baldwin (simonb@caldera.com)
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 */

%{
#include <assert.h>
#include <stdio.h>
#include <unistd.h>
#include <string.h>
#include <ctype.h>
#include <stdlib.h>
#include <malloc.h>
#include <stdarg.h>
#include <errno.h>

#include "ident.h"
#include "ifile.h"
#include "pstore.h"

#include "forth_y.h"

/* Definitions. */
#define	YYSTYPE			char*
#define	YYERROR_VERBOSE		1
#define	DESCENT_STACK_SIZE	1024
#define	MAX_ALLOWED_ERRORS	20
#define	OPERAND_BUF_LEN		4096

/* True/False macros. */
#define FALSE			(0)
#define TRUE			(!FALSE)

/*
 * Macro to cause yyparse to abort if the maximum number of allowable errors
 * has been exceeded.  This is an icky thing to have to do, but the closed
 * nature of yyparse seems to leave little option - even a setjmp/longjmp is
 * not an improvement.  And it doesn't even trigger until we hit a valid
 * token, so we may well have a lot more errors than the maximum allowed by
 * the time that happens.  All in all, rather unpleasant.
 */
#define	MNE	do {							\
			if (yy_errcount >= MAX_ALLOWED_ERRORS)		\
			    {						\
				yyerror ("too many errors, giving up");	\
				YYABORT;				\
			    }						\
		} while (0)


/* Count of errors and warnings found in the file. */
int		yy_errcount	= 0;
int		yy_warncount	= 0;

/* Flag to help us tell if we are inside yyparse, or not. */
static int	in_yyparse	= FALSE;

/* Line number from lex, and lex infile stream reset function. */
extern int	yy_linenum;
extern void	yyrestart (FILE*);

/* External declaration of yylex lexical tokenizer. */
extern int	yylex (void);

/* Forward declarations of error and parse terminal handlers. */
void		yyerror (const char *message);
static void	f_definedvariable (char *s);
static void	f_definedarray (char *s, char *ssize, char *type);
static void	f_definedconstant (char *s);
static void	f_definedword (char *s);
static void	f_semicolon (char *s);
static void	f_exit (char *s);
static void	f_pstring (char *s);
static void	f_cstring (char *s);
static void	f_word (char *s);
static void	f_object_address (char *s);
static void	f_char_literal (char *s);
static void	f_integer_literal (char *s);
static void	f_float_literal (char *s);
static void	f_string_literal (char *s);
static void	f_if (char *s);
static void	f_then (char *s);
static void	f_else (char *s);
static void	f_do (char *s);
static void	f_query_do (char *s);
static void	f_loop (char *s);
static void	f_plus_loop (char *s);
static void	f_leave (char *s);
static void	f_begin (char *s);
static void	f_until (char *s);
static void	f_again (char *s);
static void	f_while (char *s);
static void	f_repeat (char *s);
static void	f_case (char *s);
static void	f_of (char *s);
static void	f_endof (char *s);
static void	f_endcase (char *s);
static void	f_recurse (char *s);
static void	f_codeword (char *s);
static void	f_code (char *s);
static void	f_end_code (char *s);
%}

%token VARIABLE CONSTANT
%token CREATE ALLOT
%token EXIT
%token IF THEN ELSE
%token DO QUERY_DO "?DO" LOOP PLUS_LOOP "+LOOP" LEAVE
%token _BEGIN "BEGIN" UNTIL WHILE REPEAT AGAIN
%token CASE OF ENDOF ENDCASE
%token RECURSE
%token PSTRING "print string"
%token CSTRING "counted string"
%token INTEGER_LITERAL "integer"
%token STRING_LITERAL "literal string"
%token FLOAT_LITERAL "float"
%token CHAR_LITERAL "char literal"
%token OBJECT_ADDRESS "address"
%token WORD "word"
%token CODE ASM "assembler string" END_CODE "END-CODE"

%%
program:		elements
			;

elements:		/* Empty. */
			| elements element
			;

element:		declaration
			| definition
			| statement
			| error	WORD		{ MNE; yyerrok; }
			;

declaration:		VARIABLE WORD		{ MNE; f_definedvariable ($2); }
			| CONSTANT WORD		{ MNE; f_definedconstant ($2); }
			| CREATE WORD INTEGER_LITERAL WORD ALLOT { MNE;
						f_definedarray ($2, $3, $4); }
			| CREATE WORD INTEGER_LITERAL ALLOT { MNE;
						f_definedarray ($2, $3, NULL);}
			;

definition:		':' definedword statements _semicolon
			| CODE codeword codes _end_code
			;
definedword:		WORD			{ MNE; f_definedword ($1); }
			;
_semicolon:		';'			{ MNE; f_semicolon ($1); }
			;
codeword:		WORD			{ MNE; f_codeword ($1); }
			;
codes:			code
			| codes code
			;
code:			ASM			{ MNE; f_code ($1); }
			;
_end_code:		END_CODE		{ MNE; f_end_code ($1); }
			;

statement:		if_statement
			| do_loop
			| do_leave
			| begin_loop
			| begin_while
			| case_statement
			| recurse
			| EXIT			{ MNE; f_exit ($1); }
			| PSTRING		{ MNE; f_pstring ($1); }
			| CSTRING		{ MNE; f_cstring ($1); }
			| INTEGER_LITERAL	{ MNE; f_integer_literal ($1); }
			| FLOAT_LITERAL		{ MNE; f_float_literal ($1); }
			| STRING_LITERAL	{ MNE; f_string_literal ($1); }
			| CHAR_LITERAL		{ MNE; f_char_literal ($1); }
			| OBJECT_ADDRESS	{ MNE; f_object_address ($1); }
			| ALLOT			{ MNE; f_word ($1); }
			| WORD			{ MNE; f_word ($1); }
			;

if_statement:		_if statements _then
			| _if statements _else statements _then
			;
_if:			IF			{ MNE; f_if ($1); }
			;
_then:			THEN			{ MNE; f_then ($1); }
			;
_else:			ELSE			{ MNE; f_else ($1); }
			;

do_loop:		_do statements _loop
			| _do statements _plus_loop
			;
_do:			DO			{ MNE; f_do ($1); }
			| QUERY_DO 		{ MNE; f_query_do ($1); }
			;
_loop:			LOOP			{ MNE; f_loop ($1); }
			;
_plus_loop:		PLUS_LOOP		{ MNE; f_plus_loop ($1); }
			;
do_leave:		LEAVE			{ MNE; f_leave ($1); }
			;

begin_loop:		_begin statements _until
			| _begin statements _again
			| _begin statements _repeat
			;
_begin:			_BEGIN			{ MNE; f_begin ($1); }
			;
_until:			UNTIL			{ MNE; f_until ($1); }
			;
_again:			AGAIN			{ MNE; f_again ($1); }
			;
_repeat:		REPEAT			{ MNE; f_repeat ($1); }
			;
begin_while:		WHILE			{ MNE; f_while ($1); }
			;

case_statement:		_case casebody statements _endcase
			;
casebody:		caseclause
			| casebody caseclause
			;
caseclause:		statements _of statements _endof
			;
_case:			CASE			{ MNE; f_case ($1); }
			;
_endcase:		ENDCASE			{ MNE; f_endcase ($1); }
			;
_of:			OF			{ MNE; f_of ($1); }
			;
_endof:			ENDOF			{ MNE; f_endof ($1); }
			;

recurse:		RECURSE			{ MNE; f_recurse ($1); }
			;

statements:		elements
			;
%%


/*
 * do_message()
 * yyerror[a](),yywarning[a]()
 *
 * Parse error functions, also called from lexical analyzer, one
 * offering a bit more error detail, and parse warning functions.
 */
static void
do_message (const char *format, va_list ap)
{
	assert (format != NULL);

	/*
	 * If in_yyparse is false then we are not in the process of parsing
	 * a file, so there is no valid "near '...'" to print.  Otherwise,
	 * show the latest token found as a reference point in the source.
	 */
	if (in_yyparse)
		fprintf (stderr, "%s:%d: near '%s', ",
					get_ifname (), yy_linenum, yylval);
	else
		/* We are not "near" anything. */
		fprintf (stderr, "%s: ", get_ifname ());

	/* Print the error details, followed by a newline. */
	vfprintf (stderr, format, ap);
	fprintf (stderr, "\n");
}
void
yyerrora (const char *format, ...)
{
	va_list		ap;			/* Varargs list. */

	va_start (ap, format);
	do_message (format, ap);
	va_end (ap);
	yy_errcount++;
}
void
yyerror (const char *message)
{
	yyerrora ("%s", message);
}
void
yywarninga (const char *format, ...)
{
	va_list		ap;			/* Varargs list. */

	va_start (ap, format);
	do_message (format, ap);
	va_end (ap);
	yy_warncount++;
}
void
yywarning (const char *message)
{
	yywarninga ("%s", message);
}


/*
 * Parse descent (label) stack, index, and next label counter; label 0 is
 * reserved for 'main' code.
 */
static int	label_stack[DESCENT_STACK_SIZE];
static int	label_index	= 0;
static int	label_seq	= 1;

/*
 * labeltos(),labelpush(),labelpop()
 *
 * Definitions for pushing a label onto the stack, and popping one off.
 */
#ifdef FOR_FUTURE_EXPANSION
static int
labeltos (void)
{
	if (label_index <= 0)
		yyerror ("internal error: parsing stack underrun");
	return label_stack[label_index - 1];
}
#endif
static int
labelpush (void)
{
	if (label_index >= DESCENT_STACK_SIZE - 1)
		yyerror ("internal error: statement nesting too deep");
	label_stack[label_index++] = label_seq++;
	return label_stack[label_index - 1];
}
static int
labelpop (void)
{
	if (label_index <= 0)
		yyerror ("internal error: parsing stack underrun");
	return label_stack[--label_index];
}

/*
 * Note of the label number to emit at the end of the current function
 * definition - zero is special for the 'main' function.
 */
static int	end_label	= 0;

/* Pointer to the symbol table entry of any currently entered definition. */
static struct symrec	*currdef	= NULL;

/* Note if implicit 'main' code seen. */
static int	implicit_main	= FALSE;

/*
 * Break/leave stack, and current stack location/index pointer.
 */
static int	break_stack[DESCENT_STACK_SIZE];
static int	break_index	= 0;

/*
 * breaktos(),breakpush(),breakpop()
 *
 * Loop/case labels nesting stack, for leave command, and endcase break label.
 */
static int
breaktos (void)
{
	if (break_index <= 0)
		yyerror ("internal error: loop stack underrun");
	return break_stack[break_index - 1];
}
static void
breakpush (int label_number)
{
	if (break_index >= DESCENT_STACK_SIZE - 1)
		yyerror ("internal error: loop nesting too deep");
	break_stack[break_index++] = label_number;
}
static int
breakpop (void)
{
	if (break_index <= 0)
		yyerror ("internal error: loop stack underrun");
	return break_stack[--break_index];
}


/*
 * adjstr(),skipspace(),skipnonspace(),truncspace()
 *
 * General utility functions for parse handlers.
 */
static char *
adjstr (char *s)
{
	assert (s != NULL);

	/* Adjust a string by removing the initial x"space, and trailing ". */
	*(s + strlen(s) - 1) = '\0';
	return s + strlen ("x\" ");
}
static char *
skipspace (char *s)
{
	char		*sp = s;
	assert (s != NULL);

	/* Skip leading whitespace. */
	while (isspace (*sp))
		sp++;
	return sp;
}
static char *
skipnonspace (char *s)
{
	char		*sp = s;
	assert (s != NULL);

	/* Skip leading non-whitespace. */
	while (!isspace (*sp))
		sp++;
	return sp;
}
static void
truncspace (char *s)
{
	char		*sp = skipnonspace (s);
	assert (s != NULL);

	/* Truncate to the first space char found. */
	*sp = '\0';
}


/*
 * addpcode()
 *
 * Helper function to add pcode, via p_append, to the pstore.  The function
 * performs several tasks for us.  First, it detects whether this module
 * contains 'main' - the condition for containing 'main' is detecting a
 * statement outside of any definition.  Second, it finds the current
 * lexical line number, and adds that to the call to p_append.  Finally, it
 * has printf-like arguments for the pcode operand, so it gives a somewhat
 * more convenient interface to p_append for what we have to do.
 */
static void
addpcode (char *source_token, char *opcode, char *operand_format, ...)
{
	va_list		ap;			/* Varargs list. */
	char		operand[OPERAND_BUF_LEN];
	struct symrec	*symp;
	assert (opcode != NULL);

	/* Check to see if pcode is being added outside a definition. */
	if (currdef == NULL
			&& !implicit_main)
	    {
		/*
		 * The module contains an implicit 'main'.  Add the symbol
		 * for it, or complain if there's already an explicit 'main'
		 * defined.
		 */
		symp = p_findsym (mangle ("main"));
		if (symp != NULL)
			yyerrora ("'%s' both defined and implicit"
				  " (line %d)", "main", symp->linenum);
		else
			p_addsym (P_WORD, mangle ("main"), 1, P_CELL,
								yy_linenum);

		/* Set the flag that says we added an implicit 'main'. */
		implicit_main = TRUE;
	    }

	/*
	 * If operand format is not null, then there is an operand.  Other-
	 * wise, this pcode has no operand, and we don't need to format
	 * anything using sprintf.
	 */
	if (operand_format != NULL)
	    {
		/*
		 * Build the operand into a single string, and pass it to
		 * p_append.  Remember to pass on the source token, and add
		 * the line number.
		 */
		va_start (ap, operand_format);
		vsnprintf (operand, sizeof (operand), operand_format, ap);
		va_end (ap);

		p_append (opcode, operand, source_token, yy_linenum);
	    }
	else
	    {
		/*
		 * No operand, so append the pcode as above, but with the
		 * null operand field.
		 */
		p_append (opcode, NULL, source_token, yy_linenum);
	    }
}


/*
 * f_definedcommon(),f_definedvariable(),f_definedarray()
 * f_definedconstant()
 *
 * Parse terminals handlers for variable definitions.
 */
static void
f_definedcommon (char *s, int size, enum elunits units)
{
	char		*mangled_s = mangle (s);
	struct symrec	*symp = p_findsym (mangled_s);

	/* Add variable to symbol table if not yet defined. */
	if (size <= 0)
		yyerrora ("item '%s' cannot have size <= 0", s);
	else
	    {
		if (symp != NULL)
		    {
			if (symp->type != P_EXTERNALWORD)
				yyerrora ("multiple symbol definition '%s'"
					  " (line %d)", s, symp->linenum);
			else
				yyerrora
				 ("symbol '%s' already implicitly external"
					  " (line %d)", s, symp->linenum);
		    }
		p_addsym (P_VARIABLE, mangled_s, size, units, yy_linenum);
	    }
}
static void
f_definedvariable (char *s)
{
	assert (s != NULL);

	/* Define a new single cell. */
	f_definedcommon (s, 1, P_CELL);
}
static void
f_definedarray (char *s, char *ssize, char *type)
{
	assert (s != NULL && ssize != NULL);

	/* Define an array of cells/chars; NULL type implies chars. */
	truncspace (s); truncspace (ssize);
	if (type == NULL)
		f_definedcommon (s, strtol (ssize, NULL, 0), P_CHAR);
	else
	    {
		truncspace (type);
		if (strcmp (type, "CHARS") == 0)
			f_definedcommon (s, strtol (ssize, NULL, 0), P_CHAR);
		else
		    {
			if (strcmp (type, "CELLS") == 0)
				f_definedcommon (s, strtol (ssize, NULL, 0),
									P_CELL);
			else
				yyerror
				  ("parse error, expecting 'CHARS' or 'CELLS'");
		    }
	    }
}
#ifdef FOR_FUTURE_EXPANSION
static void
f_definedchararray (char *s, char *ssize)
{
	assert (s != NULL && ssize != NULL);

	/* Define an char array. */
	truncspace (s); truncspace (ssize);
	f_definedcommon (s, strtol (ssize, NULL, 0), P_CHAR);
}
#endif
static void
f_definedconstant (char *s)
{
	char		*mangled_s = mangle (s);
	struct symrec	*symp = p_findsym (mangled_s);

	/* Add constant to symbol table if not yet defined. */
	if (symp != NULL ) 
	    {
		if (symp->type != P_EXTERNALWORD)
			yyerrora ("multiple symbol definition '%s'"
				  " (line %d)", s, symp->linenum);
		else
			yyerrora ("symbol '%s' already implicitly external"
				  " (line %d)", s, symp->linenum);
	    }
	symp = p_addsym (P_CONSTANT, mangled_s, 1, P_CELL, yy_linenum);
	addpcode (s, "pop", "R0,^D");
	addpcode (NULL, "store", "R0,@%d", symp->id);
}


/*
 * f_definedword(),f_semicolon(),f_exit(),f_recurse
 *
 * Parse terminal handlers for word definition, word exit, and recurse.
 */
static void
f_definedword (char *s)
{
	char		*mangled_s = mangle (s);
	struct symrec	*symp = p_findsym (mangled_s);

	/* Add word to symbol table if not yet defined, or undefined. */
	if (symp != NULL)
	    {
		/* If undefined, change the symbol to be a defined word. */
		if (symp->type != P_EXTERNALWORD)
			yyerrora ("multiple symbol definition '%s'"
				  " (line %d)", s, symp->linenum);
		else
			p_chgsym (symp, P_WORD, 1, P_CELL);
	    }
	else
		/* Add the new symbol for this word. */
		symp = p_addsym (P_WORD, mangled_s, 1, P_CELL, yy_linenum);

	/* Set as current definition, and add pcode for a define. */
	currdef = symp;
	end_label = labelpush ();
	addpcode (s, "define", "#%d", symp->id);
}
static void
f_semicolon (char *s)
{
	assert (s != NULL);

	/* End a word definition. */
	addpcode (s, "label", "$%d", labelpop ());
	addpcode (NULL, "enddef", NULL);
	end_label = 0;
	currdef = NULL;
}
static void
f_exit (char *s)
{
	assert (s != NULL);

	/*
	 * Exit prematurely from this definition/function, via label
	 * zero if not currently inside a definition.
	 */
	addpcode (s, "jump", "$%d", end_label);
}
static void
f_recurse (char *s)
{
	assert (s != NULL);

	/* Call the routine we are already defining - for compatibility. */
	if (currdef == NULL)
		yyerrora ("'%s' invalid outside a definition", s);
	else
		addpcode (s, "call", "#%d", currdef->id);
}


/*
 * f_pstring(),f_cstring(),f_string_literal
 *
 * Parse terminal handlers for printed strings, counted strings, and string
 * literals.
 */
static void
f_pstring (char *s)
{
	char		*adjs = adjstr (s);
	struct datarec	*datap;
	char		*mangled_type = mangle ("TYPE");
	struct symrec	*symp = p_findsym (mangled_type);

	/*
	 * Add a string to data table, and emit code to print it.  First, add
	 * a symbol table definition for "TYPE" if necessary.
	 */
	if (symp != NULL)
	    {
		if (symp->type != P_EXTERNALWORD
				&& symp->type != P_WORD
				&& symp->type != P_CODEWORD)
			yyerrora ("'%s' is defined as a variable"
				  " (line %d)", "TYPE", symp->linenum);
	    }
	else
		symp = p_addsym (P_EXTERNALWORD, mangled_type, 0, P_UNKNOWN,
								yy_linenum);

	/* Now, stack data and add a call to "TYPE". */
	datap = p_adddata (P_STRING, adjs, strlen (adjs), yy_linenum);
	addpcode (s, "load", "R0,'%d", datap->id);
	addpcode (NULL, "push", "R0,^D");
	addpcode (NULL, "load", "R0,$%d", datap->len);
	addpcode (NULL, "push", "R0,^D");
	addpcode (NULL, "call", "#%d", symp->id);
}
static void
f_cstring (char *s)
{
	char		*adjs = adjstr (s);
	struct datarec	*datap;

	/* Add a string to data table, and push its address. */
	datap = p_adddata (P_COUNTEDSTRING, adjs, strlen (adjs), yy_linenum);
	addpcode (s, "load", "R0,'%d", datap->id);
	addpcode (NULL, "push", "R0,^D");
}
static void
f_string_literal (char *s)
{
	char		*adjs = adjstr (s);
	struct datarec	*datap;

	/* Add string to data table, and push its address and length. */
	datap = p_adddata (P_STRING, adjs, strlen (adjs), yy_linenum);
	addpcode (s, "load", "R0,'%d", datap->id);
	addpcode (NULL, "push", "R0,^D");
	addpcode (NULL, "load", "R0,$%d", datap->len);
	addpcode (NULL, "push", "R0,^D");
}


/*
 * f_integer_literal(),f_float_literal(),f_char_literal()
 *
 * Parse terminal handlers for integer, float, and character literal values.
 */
static void
f_integer_literal (char *s)
{
	union literal {
		int		i;
		unsigned int	u;
	} literal;
	assert (s != NULL);
	assert (sizeof (unsigned int) == sizeof (int));

	/*
	 * Push an integer literal onto the data stack.  Try this as a signed
	 * integer first, then unsigned.
	 */
	errno = 0;
	literal.i = strtol (s, NULL, 0);
	if (errno == ERANGE)
	    {
		if (*s == '-')
			yyerrora ("integer literal '%s' out of range", s);
		else
		    {
			errno = 0;
			literal.u = strtoul (s, NULL, 0);
			if (errno == ERANGE)
				yyerrora ("integer literal '%s' out of range",
									s);
		    }
	    }

	/* Append pcode to load the signed bit pattern of the literal. */
	addpcode (s, "load", "R0,$%d", literal.i);
	addpcode (NULL, "push", "R0,^D");
}
static void
f_float_literal (char *s)
{
	union literal {
		int		i;
		float		f;
	} literal;
	assert (s != NULL);
	assert (sizeof (float) == sizeof (int));

	/*
	 * Push a float literal onto the floating stack.  Convert string to
	 * a float, then push as a signed int bit pattern.
	 */
	if (sscanf (s, "%f", &literal.f) != 1)
		yyerrora ("floating literal '%s' conversion error", s);
	addpcode (s, "load", "R0,$%d", literal.i);
	addpcode (NULL, "push", "R0,^F");
}
static void
f_char_literal (char *s)
{
	char		*sp = skipspace (skipnonspace (s));

	/* Push the ASCII of the first character of s. */
	if (strlen (sp) > 1)
		yywarninga ("warning: char literal '%s' >1 character", sp);
	addpcode (s, "load", "R0,$%d", (int) *sp);
	addpcode (NULL, "push", "R0,^D");
}


/*
 * f_object_address()
 *
 * Parse terminal handler for obtaining a word's address.
 */
static void
f_object_address (char *s)
{
	char		*sp = skipspace (skipnonspace (s));
	char		*mangled_sp = mangle (sp);
	struct symrec	*symp = p_findsym (mangled_sp);

	/* Find the object, and push its address. */
	if (symp == NULL)
		symp = p_addsym (P_EXTERNALWORD, mangled_sp, 0, P_UNKNOWN,
								yy_linenum);
	addpcode (s, "load", "R0,#%d", symp->id);
	addpcode (NULL, "push", "R0,^D");
}


/*
 * f_word()
 *
 * Parse terminal handler for a word that isn't any known keyword.
 */
static void
f_word (char *s)
{
	char		*mangled_s = mangle (s);
	struct symrec	*symp = p_findsym (mangled_s);

	/*
	 * If word is a variable, push its address, if it is a constant,
	 * push its value, and if it is neither treat it as a function call.
	 * Begin by adding the symbol as an external word if it's not
	 * currently known.
	 */
	if (symp == NULL)
		symp = p_addsym (P_EXTERNALWORD, mangled_s, 0, P_UNKNOWN,
								yy_linenum);

	/* Handle the symbol according to its type. */
	switch (symp->type)
	    {
		case P_VARIABLE:
			addpcode (s, "load", "R0,#%d", symp->id);
			addpcode (NULL, "push", "R0,^D");
			break;

		case P_CONSTANT:
			addpcode (s, "load", "R0,@%d", symp->id);
			addpcode (NULL, "push", "R0,^D");
			break;

		case P_EXTERNALWORD:
		case P_WORD:
		case P_CODEWORD:
			addpcode (s, "call", "#%d", symp->id);
			break;

		default:
			yyerror ("internal error: invalid symbol type");
			break;
	    }
}


/*
 * f_if(),f_else(),f_then()
 *
 * Parse terminal handlers for IF...ELSE...THEN constructs.
 */
static void
f_if (char *s)
{
	assert (s != NULL);

	/* Test flag, and jump round if false. */
	addpcode (s, "pop", "R0,^D");
	addpcode (NULL, "jump_z", "R0,$%d", labelpush ());
}
static void
f_else (char *s)
{
	int	over = labelpop ();

	/* Jump over, then place label to jump to on if. */
	addpcode (s, "jump", "$%d", labelpush ());
	addpcode (NULL, "label", "$%d", over);
}
static void
f_then (char *s)
{
	/* Place end-of-ifthenelse label. */
	addpcode (s, "label", "$%d", labelpop ());
}


/*
 * f_do(),f_query_do(),f_leave(),f_loop(),f_plus_loop(),f_begin(),
 * f_again(),f_until(),f_while(),f_repeat()
 *
 * Parse terminal handlers for all recognized forms of loop construct.
 */
static void
f_do (char *s)
{
	/*
	 * Calculate loop iterations from the top two data stack items,
	 * and push the iterations data onto the return stack.
	 */
	addpcode (s, "pop", "R1,^D");
	addpcode (NULL, "pop", "R0,^D");
	labelpush ();
	addpcode (NULL, "push", "R0,^R");
	addpcode (NULL, "subtract", "R1,R0");
	addpcode (NULL, "push", "R1,^R");
	addpcode (NULL, "label", "$%d", labelpush ());
	breakpush (labelpush ());
}
static void
f_query_do (char *s)
{
	/* Calculate iterations as for DO, with jump over if equal. */
	addpcode (s, "pop", "R1,^D");
	addpcode (NULL, "pop", "R0,^D");
	addpcode (NULL, "jump_eq", "R0,R1,$%d", labelpush ());
	addpcode (NULL, "push", "R0,^R");
	addpcode (NULL, "subtract", "R1,R0");
	addpcode (NULL, "push", "R1,^R");
	addpcode (NULL, "label", "$%d", labelpush ());
	breakpush (labelpush ());
}
static void
f_leave (char *s)
{
	assert (s != NULL);

	/* Jump out to break label. */
	if (break_index == 0)
		yyerrora ("'%s' invalid outside a loop", s);
	else
		addpcode (s, "jump", "$%d", breaktos ());
}
static void
f_loop (char *s)
{
	int	endl = labelpop ();

	/* Test loop end, and place labels for break jump. */
	addpcode (s, "pop", "R0,^R");
	addpcode (NULL, "add", "R0,$1");
	addpcode (NULL, "push", "R0,^R");
	addpcode (NULL, "jump_nz", "R0,$%d", labelpop ());
	addpcode (NULL, "label", "$%d", endl);
	addpcode (NULL, "pop", "R0,^R");
	addpcode (NULL, "pop", "R0,^R");
	addpcode (NULL, "label", "$%d", labelpop ());
	breakpop ();
}
static void
f_plus_loop (char *s)
{
	int	endl = labelpop ();

	/* Test loop end, and place labels for break jump. */
	addpcode (s, "pop", "R1,^D");
	addpcode (NULL, "pop", "R0,^R");
	addpcode (NULL, "add", "R0,R1");
	addpcode (NULL, "push", "R0,^R");
	addpcode (NULL, "jump_ge", "R0,$%d", labelpop ());
	addpcode (NULL, "label", "$%d", endl);
	addpcode (NULL, "pop", "R0,^R");
	addpcode (NULL, "pop", "R0,^R");
	addpcode (NULL, "label", "$%d", labelpop ());
	breakpop ();
}
static void
f_begin (char *s)
{
	/* Place loop top label, and reserve break label. */
	addpcode (s, "label", "$%d", labelpush ());
	breakpush (labelpush ());
}
static void
f_again (char *s)
{
	int	endl = labelpop ();

	/* Jump to loop top, and place break label. */
	addpcode (s, "jump", "$%d", labelpop ());
	addpcode (NULL, "label", "$%d", endl);
	breakpop ();
}
static void
f_until (char *s)
{
	int	endl = labelpop ();

	/* Test stack top, jump to loop top, and place break label. */
	addpcode (s, "pop", "R0,^D");
	addpcode (NULL, "jump_z", "R0,$%d", labelpop ());
	addpcode (NULL, "label", "$%d", endl);
	breakpop ();
}
static void
f_while (char *s)
{
	assert (s != NULL);

	/* Test stack top, and if false jump to loop break. */
	if (break_index == 0)
		yyerrora ("'%s' invalid outside a loop", s);
	else
	    {
		addpcode (s, "pop", "R0,^D");
		addpcode (NULL, "jump_z", "R0,$%d", breaktos ());
	    }
}
static void
f_repeat (char *s)
{
	int	endl = labelpop ();

	/* Jump to loop top, and place loop end label. */
	addpcode (s, "jump", "$%d", labelpop ());
	addpcode (NULL, "label", "$%d", endl);
	breakpop ();
}


/*
 * f_case(),f_of(),f_endof(),f_endcase()
 *
 * Parse terminal handlers for CASE...ENDCASE constructs.
 */
static void
f_case (char *s)
{
	/* No actual code, but push an index for the endcase label. */
	addpcode (s, "no_op", NULL);
	breakpush (labelpush ());
}
static void
f_of (char *s)
{
	/* Test top against top - 1, skip over if different. */
	addpcode (s, "pop", "R1,^D");
	addpcode (NULL, "pop", "R0,^D");
	addpcode (NULL, "push", "R0,^D");
	addpcode (NULL, "jump_ne", "R0,R1,$%d", labelpush ());
	addpcode (NULL, "pop", "R0,^D");
}
static void
f_endof (char *s)
{
	/* Jump to end of case, then place skip label. */
	addpcode (s, "jump", "$%d", breaktos ());
	addpcode (NULL, "label", "$%d", labelpop ());
}
static void
f_endcase (char *s)
{
	/* Pop case value, place end of case label. */
	addpcode (s, "pop", "R0,^D");
	addpcode (NULL, "label", "$%d", labelpop ());
	breakpop ();
}


/*
 * f_codeword(),f_code(),f_endcode()
 *
 * Special parse terminal handlers for pass-through assembler (is there
 * some easier way than this?).
 */
static void
f_codeword (char *s)
{
	char		*mangled_s = mangle (s);
	struct symrec	*symp = p_findsym (mangled_s);

	/*
	 * Add word to the symbol table if not yet defined, or alter its
	 * definition if it is.  If known, but undefined, change it into
	 * a real codeword.
	 */
	if (symp != NULL)
	    {
		if (symp->type != P_EXTERNALWORD)
			yyerrora ("multiple symbol definition '%s'", s);
		else
			p_chgsym (symp, P_CODEWORD, 1, P_CELL);
	    }
	else
		/* Add a new code word symbol. */
		symp = p_addsym (P_CODEWORD, mangled_s, 1, P_CELL, yy_linenum);

	/* Set this to be the current definition, and start pcode define. */
	currdef = symp;
	addpcode (s, "define", "#%d", symp->id);
}
static void
f_code (char *s)
{
	char	*sp;
	assert (s != NULL);

	/*
	 * Add the code line to the generated code trail.  First, strip
	 * [;|#]... comments; it's very awkward to get flex to do it.
	 */
	for (sp = s; *sp != '\0'; sp++)
	    {
		/*
		 * If the current character is '"', scan forward to the
		 * next '"', or the end of line, to avoid finding ';' or
		 * '#' in strings.
		 */
		if (*sp == '"')
			for (sp++; *sp != '\0' && *sp != '"'; sp++)
				if (*sp == '\\')
					sp++;

		/* Check for a comment indicator; one of ';' or '#'. */
		if (strchr ("#;", *sp) != NULL)
		    {
			/* Terminate the string here. */
			*sp = '\0';

			/*
			 * For neatness, back up through any whitespace we
			 * find ahead of the comment indicator we just removed.
			 */
			for (sp--; sp >= s && isspace (*sp); sp--)
				*sp = '\0';
		    }
	    }

	/*
	 * Write the code verbatim, if there is anything left.  Take care to
	 * not allow '%'s in the source string to be intepolated.
	 */
	if (strlen (s) > 0)
		addpcode (NULL, "asm", "%s", s);
}
static void
f_end_code (char *s)
{
	/* End definition. */
	addpcode (s, "enddef", NULL);
	currdef = NULL;
}


/*
 * add_main_symbol()
 *
 * Helper function, adds the given string as a symbol in the main module.
 * The code generator expects to find these symbols in order to copy
 * information from the stack on process startup to them.
 */
static void
add_main_symbol (char *name)
{
	char		*mangled_name = mangle (name);
	struct symrec	*symp = p_findsym (mangled_name);

	/* If the symbol is not new or undefined, it should be. */
	if (symp != NULL)
	    {
		if (symp->type != P_EXTERNALWORD)
			yyerrora ("'%s' defined in main module"
				  " (line %d)", name, symp->linenum);
		else
			p_chgsym (symp, P_VARIABLE, 1, P_CELL);
	    }
	else
		/* Add the symbol, even though its not used in this module. */
		p_addsym (P_VARIABLE, mangled_name, 1, P_CELL, yy_linenum);
}


/*
 * yy_parse_ifile()
 *
 * Parse file entry function.
 */
int yy_parse_ifile (void) {
	int		yyparse_status = 0;

	/* Clear down states, counts, and sequences from any prior parse. */
	yy_linenum	= 1;
	yy_errcount	= 0;
	yy_warncount	= 0;

	/* Clean label sequence and stack. */
	label_index	= 0;
	label_seq	= 1;

	/* Clear definition end labels. */
	end_label	= 0;

	/* Clear any pending definitions and breaks. */
	currdef		= NULL;
	break_index	= 0;

	/* Set the input file for the lexical analyser. */
	yyrestart (get_ifstream ());

	/* Run yyparse and note the result. */
	in_yyparse	= TRUE;
	yyparse_status	= yyparse ();
	in_yyparse	= FALSE;

	/*
	 * If this is a main module, add the main module arg/env symbols,
	 * and append a 'main' end to the pstore.  Use yy_linenum - 1
	 * because since the parse is actually over, the line number count
	 * is in fact one past the last line.
	 */
	if (p_findsym (mangle ("main")) != NULL)
	    {
		/* Add the main symbols, and append the end of main label. */
		add_main_symbol ("_argc");
		add_main_symbol ("_argv");
		add_main_symbol ("_envp");
		p_append ("label", "$0", NULL, yy_linenum - 1);
	    }

	/*
	 * If no errors, perform a few housekeeping checks.  This ensures
	 * that the parse happened as we think that it should.
	 */
	if (yy_errcount == 0)
	    {
		/* Check that both our parse stacks are empty. */
		if (label_index > 0)
			yyerror ("internal error: parse stack not emptied");
		if (break_index > 0)
			yyerror ("internal error: loop stack not emptied");

		/* Check currdef and end_label finish where they started. */
		if (currdef != NULL)
			yyerror ("internal error: definition not exited");
		if (end_label != 0)
			yyerror ("internal error: definition end missed");

		/*
		 * If there's nothing in the data, symbol, or pcode areas of
		 * the pstore, and we've not mentioned any errors, then we
		 * should probably warn that the whole file contained nothing
		 * as far as we could tell.
		 */
		if (p_datastart() == NULL
				&& p_symstart() == NULL
				&& p_codestart () == NULL)
			yywarning (
				"warning: file contained no symbols or code");
	    }

	/* Return the result of running yyparse. */
	return yyparse_status;
}
