/*
 * VNPForth - Compiled native Forth for IA32 Linux systems
 * Copyright (C) 2001  Simon Baldwin (simonb@caldera.com)
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 */

#include <assert.h>
#include <string.h>
#include <stdio.h>
#include <errno.h>

#include "pstore.h"
#include "ident.h"
#include "ifile.h"

#include "optimizer.h"

/* True/False macros. */
#define FALSE			(0)
#define TRUE			(!FALSE)

/* External definitions. */
extern char		*argv0;

/* Table of optimizations. */
static const struct optrec
{
	struct opstruct
	{
		const char *opcode;	/* Opcode */
		const char *operands;	/* Operands */
	};
	const int		active;	/* Turn table entry on/off. */
	const struct opstruct	op1;	/* First operation of the pair. */
	const struct opstruct	op2;	/* Second operation of the pair. */
	const struct opstruct	repl1;	/* First replacement of the pair. */
	const struct opstruct	repl2;	/* Second replacement of the pair. */
} opttable[] = {
	/* Push a register to a stack, then pop it straight back off. */
	{ TRUE, {"push", "R0,^D"}, {"pop", "R0,^D"},
				{"no_op", NULL}, {"no_op", NULL } },
	{ TRUE, {"push", "R1,^D"}, {"pop", "R1,^D"},
				{"no_op", NULL}, {"no_op", NULL } },
	{ TRUE, {"push", "R0,^R"}, {"pop", "R0,^R"},
				{"no_op", NULL}, {"no_op", NULL } },
	{ TRUE, {"push", "R1,^R"}, {"pop", "R1,^R"},
				{"no_op", NULL}, {"no_op", NULL } },
	{ TRUE, {"push", "R0,^F"}, {"pop", "R0,^F"},
				{"no_op", NULL}, {"no_op", NULL } },
	{ TRUE, {"push", "R1,^F"}, {"pop", "R1,^F"},
				{"no_op", NULL}, {"no_op", NULL } },

	/* Push one register to a stack, then pop other off the same stack. */
	{ TRUE, {"push", "R0,^D"}, {"pop", "R1,^D"},
				{"load", "R1,R0"}, {"no_op", NULL } },
	{ TRUE, {"push", "R1,^D"}, {"pop", "R0,^D"},
				{"load", "R0,R1"}, {"no_op", NULL } },
	{ TRUE, {"push", "R0,^R"}, {"pop", "R1,^R"},
				{"load", "R1,R0"}, {"no_op", NULL } },
	{ TRUE, {"push", "R1,^R"}, {"pop", "R0,^R"},
				{"load", "R0,R1"}, {"no_op", NULL } },
	{ TRUE, {"push", "R0,^F"}, {"pop", "R1,^F"},
				{"load", "R1,R0"}, {"no_op", NULL } },
	{ TRUE, {"push", "R1,^F"}, {"pop", "R0,^F"},
				{"load", "R0,R1"}, {"no_op", NULL } },

	{ TRUE, {NULL, NULL}, {NULL, NULL},
				{NULL, NULL}, {NULL, NULL} }
};


/*
 * optimize_scan1()
 *
 * Minimal optimization of pcode.  This function looks at consecutive pcode
 * operations, and tries to either eliminate operations, or to combine the
 * operations into a single replacement operation.  While there's almost no
 * limit to what we _could_ do, in practice, we'll restrict ourselves to
 * just removing unnecessary push/pop calls, which are relatively expensive,
 * and not worry at all about unnecessary register moves, jump merging, and
 * so on.
 */
static int
optimize_scan1 (void)
{
	struct pcoderec		*pcodep1, *pcodep2;
						/* Pair of pcode pointers */
	const struct optrec	*optp;		/* Optimization table pointer */
	int			optimizations = 0;
						/* Count of optimizations */

	/* Run through each pcode entry consecutive pair. */
	for (pcodep1 = p_codestart (); pcodep1 != NULL;
						pcodep1 = p_codenext (pcodep1))
	    {
		/* Find the next opcode that isn't a no_op. */
		pcodep2 = p_codenext (pcodep1);
		while (pcodep2 != NULL
				&& strcmp (pcodep2->opcode, "no_op") == 0)
			pcodep2 = p_codenext (pcodep2);
		if (pcodep2 == NULL)
			break;

		/* Scan for this pair in the replacements table. */
		for (optp = opttable;
			optp->op1.opcode != NULL && optp->op1.operands != NULL;
		       							optp++)
		    {
			/* If not active, ignore this entry. */
			if (!optp->active)
				continue;

			/* Compare the opcodes/operands with this entry. */
			if (	   !strcmp (pcodep1->opcode, optp->op1.opcode)
				&& !strcmp (pcodep1->operands,
							optp->op1.operands)
				&& !strcmp (pcodep2->opcode, optp->op2.opcode)
				&& !strcmp (pcodep2->operands,
				       			optp->op2.operands))
			    {
				/* Replace these opcodes. */
				p_replace (pcodep1, optp->repl1.opcode,
							optp->repl1.operands);
				p_replace (pcodep2, optp->repl2.opcode,
							optp->repl2.operands);
				optimizations++;
				break;
			    }
		    }
	    }

	/* Return the count of optimizations made. */
	return optimizations;
}


/*
 * warning()
 *
 * Print a warning.  The information printed is like that of yywarning, but
 * we can't actually use yywarning because parsing is finished.
 */
static void
warning (int linenum, const char *source_token, const char *message)
{
	assert (message != NULL);
	if (source_token != NULL)
		fprintf (stderr, "%s:%d: near '%s', %s\n",
				get_ifname (), linenum, source_token, message);
	else
		fprintf (stderr, "%s:%d: %s\n",
				get_ifname (), linenum, message);
}


/*
 * unreachable_scan()
 *
 * Scan for unreachable code fragments.  Here, we look through the pcode for
 * unconditional jumps.  If we find one, we then look forwards to the next
 * opcode that is not a no_op.  If the thing we find is a label, the code
 * is reachable; if not, the opcode that we found is not reachable.
 */
static void
unreachable_scan (void)
{
	struct pcoderec		*pcodep1, *pcodep2;
						/* Pair of pcode pointers */

	/* Run through each pcode entry singly, finding unconditional jumps. */
	for (pcodep1 = p_codestart (); pcodep1 != NULL;
						pcodep1 = p_codenext (pcodep1))
	    {
		/* If not an unconditional jump, ignore this entry. */
		if (strcmp (pcodep1->opcode, "jump") != 0)
			continue;

		/* Look ahead to the next opcode that isn't a no_op. */
		pcodep2 = p_codenext (pcodep1);
		while (pcodep2 != NULL
				&& strcmp (pcodep2->opcode, "no_op") == 0)
			pcodep2 = p_codenext (pcodep2);
		if (pcodep2 == NULL)
			break;

		/* If we did not stop on a label, this code is unreachable. */
		if (strcmp (pcodep2->opcode, "label") != 0)
			warning (pcodep2->linenum, pcodep2->source_token,
				 		"warning: unreachable code");
	    }
}


/*
 * optimize()
 *
 * Main call in to the pcode optimizer.
 */
void
optimize (void)
{
	/*
	 * Optimize out unnecessary code as far as possible (which isn't in
	 * fact all that far).
	 */
	while (optimize_scan1 () > 0)
		;

	/* Scan for unreachable code. */
	unreachable_scan ();
}
